#pragma OPENCL EXTENSION cl_khr_byte_addressable_store : enable
#pragma OPENCL EXTENSION cl_khr_local_int32_base_atomics : enable

#define u32 uint
#define u8 uchar
#define size_t uchar
#define rol(value, bits) (((value) << (bits)) | ((value) >> (32 - (bits))))

/* blk0() and blk() perform the initial expand. */
/* I got the idea of expanding during the round function from SSLeay */
#define blk0(i) (block->l[i] = (rol(block->l[i], 24) & 0xFF00FF00) | (rol(block->l[i], 8) & 0x00FF00FF))
#define blk(i) (block->l[i & 15] = rol(block->l[(i + 13) & 15] ^ block->l[(i + 8) & 15] ^ block->l[(i + 2) & 15] ^ block->l[i & 15], 1))

/* (R0+R1), R2, R3, R4 are the different operations used in SHA1 */
#define R0(v,w,x,y,z,i) z += ((w & (x ^ y)) ^ y) + blk0(i) + 0x5A827999 + rol(v, 5); w = rol(w, 30);
#define R1(v,w,x,y,z,i) z += ((w & (x ^ y)) ^ y) + blk(i) + 0x5A827999 + rol(v, 5); w = rol(w, 30);
#define R2(v,w,x,y,z,i) z += (w ^ x ^ y) + blk(i) + 0x6ED9EBA1 + rol(v, 5); w = rol(w, 30);
#define R3(v,w,x,y,z,i) z += (((w | x) & y) | (w & x)) + blk(i) + 0x8F1BBCDC + rol(v, 5); w = rol(w, 30);
#define R4(v,w,x,y,z,i) z += (w ^ x ^ y) + blk(i) + 0xCA62C1D6 + rol(v, 5); w=rol(w, 30);


//#define get_global_id(a) 0

#define SHA1_MAC_LEN 20

typedef local struct _ctxarray {
    uint a[128];
    uint b[128];
    uint c[128];
    uint d[128];
    uint e[128];
    uint l1[128];
} ctxarray; 

typedef union {
	unsigned char c[64];
	unsigned int l[16];
} CHAR64LONG16;


constant const size_t _len2[2] = { 64, SHA1_MAC_LEN};
constant const size_t _len21[2] = { 64, 4};
constant const char reflector[64] = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcedfghijklmnopqrstuvwxyz?@";


/* compare the private memory block to the constant memory block
 * dest: the private memory block to compare
 * src: the constant memory block to compare
 * amount: the amount of bytes to compare
 * constoffset: the offset on the const block.
 **/

void os_memcpy(private uchar* dest, private const uchar* src, const uchar amount) {
    uchar j = 0;
    for(j = 0; j < amount; j++) dest[j] = src[j];
}

/* if the global id is (0,0,0), this program will dump the output in both uchar arrays EXACTLY ONCE
 * subsequent function calls will be ignored
 * it dumps: 
 * 0. function blocker (0xff)
 * 1. last 32 address bits of array
 * 2. last 32 address bits of array2
 * 3. array[0] | array[1] << 8 | array[2] << 16 | array[3] << 24
 * 4. array2[0] | array2[1] << 8 | array2[2] << 16 | array2[3] << 24
*/
void dump_global_output(const uchar* array, const uchar* array2, const int outputoffset, global uint* output) {
	 if((get_global_id(0) != 0)|get_global_id(1)|get_global_id(2)|output[outputoffset]) return;
	output[outputoffset] = 0xff;
	output[outputoffset + 1] = (unsigned int) array;
	output[outputoffset + 2] = (unsigned int) array2;
	output[outputoffset + 3] = array[0] | array[1] << 8 | array[2] << 16 | array[3] << 24;
	output[outputoffset + 4] = array2[0] | array2[1] << 8;
}


void os_memcpyc(private void* dest, constant const void* src, const uchar amount) {
    uchar j = 0, intamount = amount >> 2;
    int* destination = (int*) dest;
    constant int* source = (constant int*) src;
    char* destination2 = (char*) dest;
    constant char* source2 = (constant char*) src;
    while(j < intamount) {
        destination[j] = source[j];
        j++;
    }
    j = j << 2;
    while(j < amount) {
        destination2[j] = source2[j];
        j++;
    }
}
void os_memcpyzero(private void* dest, const uchar amount) {
    uchar j = 0;
    char* destination2 = (char*) dest;
    while(j < amount) {
        destination2[j] = 0;
        j++;
    }
}

void os_memcpywithIDs(private void* dest, constant const void* src, const uchar amount, const uint idmap) {
    uchar j;
    uint* dst = (uint*) dest;
    constant uint* source = (constant uint*) src;
    for(j = (amount >> 2) - 1; j > 0; j--) dst[j] = source[j];
    dst[0] = idmap;
}
	
/* Hash a single 512-bit block. This is the core of the algorithm. */

void SHA1Transform(private unsigned char buffer[64], ctxarray* ctx, global uint* output)
{
	unsigned int a, b, c, d, e;
	CHAR64LONG16* block;
	block = (CHAR64LONG16 *) buffer;
	/* Copy context->state[] to working vars */
	a = ctx->a[get_local_id(0)];
	b = ctx->b[get_local_id(0)];
	c = ctx->c[get_local_id(0)];
	d = ctx->d[get_local_id(0)];
	e = ctx->e[get_local_id(0)];
	/* 4 rounds of 20 operations each. Loop unrolled. */
	R0(a,b,c,d,e, 0); R0(e,a,b,c,d, 1); R0(d,e,a,b,c, 2); R0(c,d,e,a,b, 3);
	R0(b,c,d,e,a, 4); R0(a,b,c,d,e, 5); R0(e,a,b,c,d, 6); R0(d,e,a,b,c, 7);
	R0(c,d,e,a,b, 8); R0(b,c,d,e,a, 9); R0(a,b,c,d,e,10); R0(e,a,b,c,d,11);
	R0(d,e,a,b,c,12); R0(c,d,e,a,b,13); R0(b,c,d,e,a,14); R0(a,b,c,d,e,15);
	R1(e,a,b,c,d,16); R1(d,e,a,b,c,17); R1(c,d,e,a,b,18); R1(b,c,d,e,a,19);
	R2(a,b,c,d,e,20); R2(e,a,b,c,d,21); R2(d,e,a,b,c,22); R2(c,d,e,a,b,23);
	R2(b,c,d,e,a,24); R2(a,b,c,d,e,25); R2(e,a,b,c,d,26); R2(d,e,a,b,c,27);
	R2(c,d,e,a,b,28); R2(b,c,d,e,a,29); R2(a,b,c,d,e,30); R2(e,a,b,c,d,31);
	R2(d,e,a,b,c,32); R2(c,d,e,a,b,33); R2(b,c,d,e,a,34); R2(a,b,c,d,e,35);
	R2(e,a,b,c,d,36); R2(d,e,a,b,c,37); R2(c,d,e,a,b,38); R2(b,c,d,e,a,39);
	R3(a,b,c,d,e,40); R3(e,a,b,c,d,41); R3(d,e,a,b,c,42); R3(c,d,e,a,b,43);
	R3(b,c,d,e,a,44); R3(a,b,c,d,e,45); R3(e,a,b,c,d,46); R3(d,e,a,b,c,47);
	R3(c,d,e,a,b,48); R3(b,c,d,e,a,49); R3(a,b,c,d,e,50); R3(e,a,b,c,d,51);
	R3(d,e,a,b,c,52); R3(c,d,e,a,b,53); R3(b,c,d,e,a,54); R3(a,b,c,d,e,55);
	R3(e,a,b,c,d,56); R3(d,e,a,b,c,57); R3(c,d,e,a,b,58); R3(b,c,d,e,a,59);
	R4(a,b,c,d,e,60); R4(e,a,b,c,d,61); R4(d,e,a,b,c,62); R4(c,d,e,a,b,63);
	R4(b,c,d,e,a,64); R4(a,b,c,d,e,65); R4(e,a,b,c,d,66); R4(d,e,a,b,c,67);
	R4(c,d,e,a,b,68); R4(b,c,d,e,a,69); R4(a,b,c,d,e,70); R4(e,a,b,c,d,71);
	R4(d,e,a,b,c,72); R4(c,d,e,a,b,73); R4(b,c,d,e,a,74); R4(a,b,c,d,e,75);
	R4(e,a,b,c,d,76); R4(d,e,a,b,c,77); R4(c,d,e,a,b,78); R4(b,c,d,e,a,79);
	/* Add the working vars back into context.state[] */
	ctx->a[get_local_id(0)] += a;
	ctx->b[get_local_id(0)] += b;
	ctx->c[get_local_id(0)] += c;
	ctx->d[get_local_id(0)] += d;
	ctx->e[get_local_id(0)] += e;
}

/* SHA1Init - Initialize new context */

void SHA1Init(ctxarray* ctx)
{
	/* SHA1 initialization constants */
	ctx->a[get_local_id(0)] = 0x67452301;
    ctx->b[get_local_id(0)] = 0xEFCDAB89;
    ctx->c[get_local_id(0)] = 0x98BADCFE;
    ctx->d[get_local_id(0)] = 0x10325476;
    ctx->e[get_local_id(0)] = 0xC3D2E1F0;
     ctx->l1[get_local_id(0)] = 0;
}

void SHA1Update(private const unsigned char* data, private const uchar len, ctxarray* ctx, private uchar* ctxbuffer, global uint* output)
{
	ushort i, j;
	 ctx->l1[get_local_id(0)] += (len << 3);
	j = ((ctx->l1[get_local_id(0)] >> 3) - len) & 63;
	if(((j + len) & 64)) {
		os_memcpy(&ctxbuffer[j], data, (i = 64-j));
		SHA1Transform((private uchar*)ctxbuffer, ctx, output);
		j = 0;
	}
	else i = 0;
	os_memcpy(&ctxbuffer[j], &data[i], len - i);
}

void SHA1Updatezero(private const uchar len, ctxarray* ctx, private uchar* ctxbuffer)
{
	
	ushort j = (ctx->l1[get_local_id(0)] >> 3) & 63;
	 ctx->l1[get_local_id(0)] += (len << 3);
	os_memcpyzero(&ctxbuffer[j], len);
}

void SHA1Updatec(constant unsigned char* data, const uint len, ctxarray* ctx, private uchar* ctxbuffer, global uint* output)
{
	ushort i,j;
	j = (ctx->l1[get_local_id(0)] >> 3) & 63;
	 ctx->l1[get_local_id(0)] += (len << 3);
   	if (((j + len) & 64)) {
		os_memcpyc((private uchar*) &ctxbuffer[j], data, (i = 64-j));
		SHA1Transform((private uchar*)ctxbuffer, ctx, output);
		j = 0;
	}
	else {
		i = 0;
	}
	os_memcpyc((private uchar*) &ctxbuffer[j], &data[i], len - i);
}


/* Add padding and return the message digest. */

void SHA1Final(private unsigned char digest[20], ctxarray* ctx, private uchar* ctxbuffer, global uint* output)
{
	unsigned char finalcount[4];
        finalcount[0] = (unsigned char) (ctx->l1[get_local_id(0)] >> 8) & 255;  /* Endian independent */
        finalcount[1] = (unsigned char) (ctx->l1[get_local_id(0)]) & 255;  /* Endian independent */
        SHA1Updatec((constant unsigned char *) "\200", 1, ctx, ctxbuffer, output);
        SHA1Updatezero(((496 - (ctx->l1[get_local_id(0)] & 504)) >> 3), ctx, ctxbuffer);
        SHA1Update(finalcount, 2, ctx, ctxbuffer, output);  /* Should cause a SHA1Transform() */
        digest[0] = (unsigned char)(ctx->a[get_local_id(0)]  >> 24);
        digest[1] = (unsigned char)(ctx->a[get_local_id(0)]  >> 16) & 255;
        digest[2] = (unsigned char)(ctx->a[get_local_id(0)]  >> 8) & 255;
        digest[3] = (unsigned char)(ctx->a[get_local_id(0)]) & 255;
        digest[4] = (unsigned char)(ctx->b[get_local_id(0)]  >> 24);
        digest[5] = (unsigned char)(ctx->b[get_local_id(0)]  >> 16) & 255;
        digest[6] = (unsigned char)(ctx->b[get_local_id(0)]  >> 8) & 255;
        digest[7] = (unsigned char)(ctx->b[get_local_id(0)]) & 255;
        digest[8] = (unsigned char)(ctx->c[get_local_id(0)]  >> 24);
        digest[9] = (unsigned char)(ctx->c[get_local_id(0)]  >> 16) & 255;
        digest[10] = (unsigned char)(ctx->c[get_local_id(0)]  >> 8) & 255;
        digest[11] = (unsigned char)(ctx->c[get_local_id(0)]) & 255;
        digest[12] = (unsigned char)(ctx->d[get_local_id(0)]  >> 24);
        digest[13] = (unsigned char)(ctx->d[get_local_id(0)]  >> 16) & 255;
        digest[14] = (unsigned char)(ctx->d[get_local_id(0)]  >> 8) & 255;
        digest[15] = (unsigned char)(ctx->d[get_local_id(0)]) & 255;
        digest[16] = (unsigned char)(ctx->e[get_local_id(0)]  >> 24);
        digest[17] = (unsigned char)(ctx->e[get_local_id(0)]  >> 16) & 255;
        digest[18] = (unsigned char)(ctx->e[get_local_id(0)]  >> 8) & 255;
        digest[19] = (unsigned char)(ctx->e[get_local_id(0)]) & 255;

}

void sha1_vector(const size_t num_elem2, private const u8 *addr2[], constant const size_t* len2, const size_t num_elem, constant u8 *addr[], const size_t *len, private u8 *mac, global uint* output, ctxarray* ctx)
{
    	private uchar ctxbuffer[64];
	SHA1Init(ctx);
	os_memcpyzero(ctxbuffer, 64);
    	SHA1Update((private const u8*)addr2[0], len2[0], ctx, ctxbuffer, output);
        if(num_elem)
		SHA1Updatec((constant u8*)addr[0], len[0], ctx, ctxbuffer, output);
	if(num_elem2 == 2)
		SHA1Update((private const u8*)addr2[1], len2[1], ctx, ctxbuffer, output);
	SHA1Final(mac, ctx, ctxbuffer, output);

}

/**
 * hmac_sha1_vector - HMAC-SHA1 over data vector (RFC 2104)
 * @key: Key for HMAC operations
 * @key_len: Length of the key in bytes
 * @num_elem: Number of elements in the data vector
 * @addr: Pointers to the data areas
 * @len: Lengths of the data blocks
 * @mac: Buffer for the hash (20 bytes)
 * Returns: 0 on success, -1 on failure
 */
void hmac_sha1_vector(constant const char *key, const size_t key_len, const size_t num_elem, const size_t num_elem2, constant u8 **addr, private const u8 *addr2, const size_t *len, constant const size_t *len2, private u8 *mac, const uint idmap, global uint* 
output, ctxarray* ctx)
{
	private CHAR64LONG16 k_pad; /* padding - key XORd with ipad/opad */
	constant u8 *_addr[2];
    	private const u8 *_addr2[4] = { k_pad.c, addr2, k_pad.c, mac};
	size_t i;
    	constant const size_t *_lenl = len2;

	/* the HMAC_SHA1 transform looks like:
	 *
	 * SHA1(K XOR opad, SHA1(K XOR ipad, text))
	 *
	 * where K is an n byte key
	 * ipad is the byte 0x36 repeated 64 times
	 * opad is the byte 0x5c repeated 64 times
	 * and text is the data being protected */

	/* start out by storing key in ipad */
	os_memcpywithIDs(k_pad.l, key, key_len, idmap);
    	for(i = key_len >> 2; i < 16; i++) k_pad.l[i] = 0;
	/* XOR key with ipad values */
	for (i = 0; i < 16; i++)
		k_pad.l[i] ^= 0x36363636;
	/* perform inner SHA1 */
	for (i = 0; i < num_elem; i++) {
		_addr[i] = addr[i];
	}
	sha1_vector(1 + num_elem2, _addr2, _lenl, num_elem, _addr, len, mac, output, ctx);
	os_memcpywithIDs(k_pad.l, key, key_len, idmap);
	for(i = key_len >> 2; i < 16; i++) k_pad.l[i] = 0;
    /* XOR key with opad values */
	for (i = 0; i < 16; i++)
		k_pad.l[i] ^= 0x5c5c5c5c;

	sha1_vector(2, &_addr2[2], _len2, 0, _addr, len, mac, output, ctx);
}


void pbkdf2_sha1_f(constant const char *passphrase, const size_t passphrase_len, constant uchar *ssid, const size_t ssid_len, const uchar count, private u8 *digest, global uint* output, ctxarray* ctx)
{
	private unsigned char _tmp[20];
	private unsigned char _tmp2[20];
	private unsigned char *tmp = _tmp, *tmp2 = _tmp2;
	ushort i, j;
	private uchar count_buf[4];
        private const uint idmap = (reflector[(get_global_id(0) & 63)] | (reflector[(get_global_id(1) & 63)] << 8) | (reflector[(get_global_id(2) & 63)] << 16) | (reflector[((get_global_id(0) >> 6) | ((get_global_id(1) >> 4) & 0xC) | ((get_global_id(2) >> 2) & 0x30))] << 24));


	/* F(P, S, c, i) = U1 xor U2 xor ... Uc
	 * U1 = PRF(P, S || i)
	 * U2 = PRF(P, U1)
	 * Uc = PRF(P, Uc-1)
	 */

	count_buf[0] = 0;
	count_buf[1] = 0;
	count_buf[2] = 0;
	count_buf[3] = count;
	hmac_sha1_vector(passphrase, passphrase_len, 1, 1, &ssid, count_buf, &ssid_len, _len21, tmp, idmap, output, ctx);
    	os_memcpy(digest, tmp, SHA1_MAC_LEN);
	for (i = 1; i < 4096; i++) {
		hmac_sha1_vector(passphrase, passphrase_len, 0, 1, &ssid, tmp, &ssid_len, _len2, tmp2, idmap, output, ctx);
		if(tmp == _tmp) {
			tmp = _tmp2;
			tmp2 = _tmp;
		} else {
			tmp = _tmp;
			tmp2 = _tmp2;
		}
		for (j = 0; j < SHA1_MAC_LEN; j++)
			digest[j] ^= tmp[j];
	}
}

__kernel void startCrack(__constant uchar* hash, __constant uchar* ssid, const size_t ssid_len, __constant char* passphrase, const size_t key_len, __global uint* output) {
    private uchar system[4];
    private uchar enabled = 1, iterator = 0;
    ctxarray ctx;
    private unsigned char digest[20] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
	system[0] = 0;
	system[2] = 32;
    if(output[3] == 255) enabled = 0;
     while(system[2] && enabled) {
			system[0]++;
		        system[3] = system[2] > SHA1_MAC_LEN ? SHA1_MAC_LEN : system[2];
			system[1] = 32 -  system[2];
                pbkdf2_sha1_f(passphrase, key_len, ssid, ssid_len, system[0], digest, output, &ctx);
		for(iterator = 0; enabled && iterator < system[3]; iterator++) enabled = (hash[iterator+system[1]] == digest[iterator]);
                system[2] -= system[3];
    }
    if(enabled) {
        output[3] = 255;
       	output[0] = get_global_id(0);
        output[1] = get_global_id(1);
        output[2] = get_global_id(2);
    }
}


